"""Classes related to parameterizing ETLs.

"""
from .io import In, Out
from .util import timeit
import sys
import json


class Parameter:
    """
    Base class for a parameter.

    Attributes:
        name (string) parameter name
        default (string) default value
        type (string or JSONType) type of parameter value
        custom (list) list of parameter choices that will add custom parameters
        kwargs (dict) keyword arguments supplied at initialization
    """

    type = None
    custom = []

    def __init__(self, name, **kwargs):
        """

        Args:
            name (string): parameter name
            default (string) default value
            type (string or the value JSONType) type of parameter value
            custom (list) list of parameter choices that will add custom parameters
        """
        self.name = name
        self.__dict__.update(kwargs)
        self.kwargs = kwargs

    def as_injected(self, injected_type):
        """
        Method to cast this parameter as dependency injected type

        Args:
            injected_type (type) type to cast to
        """

        # collect attributes without custom
        d = {k: v for k, v in self.__dict__.items() if k != 'custom'}

        # create and return
        return injected_type(**d)


class ParameterizingOut(Out):
    """
    Class used at initialization of a parameterized object to gather parameters from its ETL and make them concrete.
    Order of parameter value precedence: cmd-line overrides > script defaults > ETL defaults.

    Attributes:
        parameterized (class): the parameterized script object with an ETL factory
        Parameter (class): the dependency injected Parameter type
        JSONType (class): the dependency injected JSONType type

    Private Attributes:
        _in (object): the intermediate state
    """

    def __init__(self, cls, Parameter, JSONType):
        self.parameterized = cls
        self.Parameter = Parameter
        self.JSONType = JSONType
        self._in = In()

    def read_parameter_values(self, args=sys.argv[2:]):
        self.read_parameter_defaults()
        self.read_parameter_overrides(args)
        self.parse_sweep()

    def read_parameter_defaults(self):
        """
        Method to read default parameter values from parameterized class into this output

        """

        # get default paremter values from parameterized class
        for var in dir(self.parameterized):
            if var[0] == '_':
                continue
            # try:
            val = getattr(self.parameterized, var)
            # except Exception:
            #     continue
            if isinstance(val, self.Parameter):
                setattr(self, var, val.kwargs['default'])

    def read_parameter_overrides(self, args=sys.argv[2:]):
        """
        Method to read overridden parameters from command line arguments

        Args:
            args (string) command line arguments
        """

        # get parameter value overrides from cmd line
        if len(args) > 0:
            if args[0] == "--help":
                print("")
                print('\033[92mRemember certain parameter choices will add new parameters!\033[0m')
                print("")
                args = args[1:]

            if not (args[1] == 'run' or (len(args) > 1 and args[0] == 'run')):
                # read from (auto-) tags
                for a, b in zip(args[::2], args[1::2]):
                    if a == "--tag":
                        ls = b.split(':')
                        setattr(self, ls[0], ":".join(ls[1:]))
            else:
                # read from non tags
                for a, b in zip(args[::2], args[1::2]):
                    if a != "--tag":
                        setattr(self, a[2:], b)

    def parse_sweep(self, sweep_json=None):
        """
        Method parse sweep parameter from json

        Args:
            sweep_json (string) the json representation of the sweep dic
        """
        if sweep_json is None:
            sweep_json = getattr(self, 'sweep', '{}')
        self.sweep = json.loads(sweep_json)

        # remove swept parameters but add first value
        for name, vals in self.sweep.items():
            # remove parameter from class
            if hasattr(self.parameterized, name):
                setattr(self.parameterized, name, "swept")
            # add first value in sweep to this output
            setattr(self, name, vals[0])

    def add_parameter(self, name, param):
        """
        Method to add a parameter to the parameterized class

        Args:
            name (string) name of parameter
            param (object) abstract parameter
        """

        # parameterized class
        cls = self.parameterized

        # dependency injection of JSONType
        if (param.type == JSONType):
            param.type = self.JSONType

        # only add parameter if it's new and not part of a sweep
        if not hasattr(cls, name) and (not hasattr(self, 'sweep') or name not in self.sweep):

            # add parameter to cls using dependency injected Parameter type
            setattr(cls, name, param.as_injected(self.Parameter))

        # handle parameter sweep
        if name == 'sweep':

            # we already have a sweep - do it only once
            if hasattr(self, "sweep"):
                return

            # parse sweep into dict
            self.parse_sweep(param.default)
        else:
            # keep parameter default param as attribute
            if not hasattr(self, name):
                setattr(self, name, param.default)

            # keep parameter in _in
            setattr(self._in, name, param)


class Parameterized:
    """
    Mixin class for parameterizing an object using its ETL.

    Assertions about mixer:
        - has attribute factories (dict) with a factories["get_etl"] factory for creating an ETL object
        - injects dependencies for implementing Parameter and JSONType via _Parameter and _JSONType attributes

    Attributes:
        sweep_index (int): Current index for parameter sweeps.
        _Parameter (type): dependency injected class for parameters
        _JSONType (type): dependency injected class for JSON parameter values
    """
    factories = {}
    doc = "A parameterized script"

    def __init__(self, **kwargs):
        self.parameterize()
        self.add_parameter_tags()
        self.sweep_index = 0
        super().__init__(**kwargs)

    def parameterize(self):
        """Method to recursively add any parameters from the ETL.

        Returns: the class for convenience
        """
        # modifications will happen to the class
        cls = self.__class__

        # the factory for creating the ETL object
        get_etl = self.factories["get_etl"]

        # create own factories dict
        cls.factories = {"get_etl": get_etl}

        # create own doc string
        cls.__doc__ = '\033[92m%s\033[0m' % self.doc

        # here we just need a foo input object
        i = In()

        # create out for adding parameters while using dependency injections for Parameter and JSONType classes
        o = ParameterizingOut(cls, self._Parameter, self._JSONType)

        # read parameter values using cls itself
        o.read_parameter_values()

        # add additional parameters recursively by simply creating the ETL object
        cls.etl = get_etl(i, o)

        # add any custom parameters for all parameter sweep cases
        for var, vals in o.sweep.items():
            for val in vals:
                if val in getattr(o._in, var).custom:
                    # fetch again with custom value
                    setattr(o, var, val)
                    get_etl(i, o)
        return cls

    def add_parameter_tags(self, tag="--tag", fmt="%s:%s", ignore=[
                'ep', 'environment', 'metadata', 'datastore', 'run-id',
                'task-id', 'event-logger', 'monitor', 'datastore-root', 'input-paths'], args=sys.argv):
        """
        Method to add parameters and their values as command line arguments e.g. in the form: --tag var:val
        The primary use case is for auto-tagging metaflow runs.

        Arguments:
        tag (str) command line argument string
        fmt (str) format for parameter name and value e.g. "%s:%s" for var:val
        ignore (list) names of command line arguments to ignore when auto-tagging
        """

        # only add tags if we run the metaflow
        if len(args) > 1 and (args[1] == 'run' or (len(args) > 3 and args[3] == 'run')):
            done = {}
            oargs = args[2:]

            # add parameters overrides from the command line arguments
            for a, b in zip(oargs[::2], oargs[1::2]):
                if a != tag and a[2:] not in ignore:
                    args.extend([tag, fmt % (a[2:], b)])
                    done[a[2:]] = True

            # add the rest of the parameters from the class itself
            for k, v in self._get_parameters():
                # skip if overridden
                if k in done:
                    continue
                # otherwise use default value
                args.extend([tag, fmt % (k, str(v.kwargs['default']))])

    def process(self, i):
        """
        Method to process a single input or a list of [i,o] pairs

        Arguments:
            i (object) a single input or a list of [i,o] pairs
        """

        # the ETL factory
        get_etl = self.factories["get_etl"]

        if isinstance(i, list):
            # decorate list with Inputs
            i = Inputs(i)

            # process i,o pairs
            for io in i:
                _i, o = io

                # TODO make it right
                # ignore parameters
                o.add_parameter = lambda i, o: 0
                o.timeit = timeit

                # create and append ETL object to list [i,o] -> [i,o,etl]
                io.append(get_etl(_i, o))

        # simple logging - TODO make it right
        print('\033[94mDo ETL %s\033[0m' % str(i))

        # create ETL and execute it using self as output
        o = self
        get_etl(i, o).etl(i, o)

    def _get_parameters(self):
        """Provide own parameters as a generator of name,value pairs.

        """
        # go through attributes by name
        for name in dir(self):
            # skip private
            if name[0] == '_':
                continue

            # make sure it is accessible and get attribute value
            # try:
            value = getattr(self, name, None)
            # except Exception:
            #    continue

            # if value is a parameter yield name,value
            if isinstance(value, self._Parameter):
                yield name, value


class Inputs(list):
    """
    Decorator class for list of [i,o] pairs
    """
    pass


class JSONType:
    """
    Abstract class for JSON type for parameter values.

    """
    pass



"""Functions for timing of method calls.

"""
import time


def tim(method):
    """
    Decorator to time a method call of an ETL if its output object supports it

    Arguments:
        method (function) the method to time
    """
    def tm(*args, **kwargs):
        slf = args[0]
        if hasattr(slf.o, "timeit"):
            return slf.o.timeit(method, slf.o, *args, **kwargs)
        else:
            return method(*args, **kwargs)
    return tm


def timeit(method, o, *args, **kw):
    """
    Time a method call

    Arguments:
        method (function) the method to time
        o (object) the output to add the result to
        *args forward remaing arguments to method
        *kw forward keyword argument to method
    """
    ts = time.time()
    result = method(*args, **kw)
    te = time.time()
    delta = int((te - ts) * 1000)
    if 'log_time' in kw:
        # explicit structuring of timings
        name = kw.get('log_name', method.__name__.upper())
        kw['log_time'][name] = delta
    else:
        # simple naming scheme: ms_class_method
        setattr(o, 'ms_%s_%s' % (type(args[0]).__name__, method.__name__),  delta)
    return result



import pytest
import json
from paretl import ParameterizingOut, Parameter, JSONType, Parameterized, ETL, In, Out, timeit, tim, Sweep


@pytest.fixture()
def par():
    return Parameter('name', default="foo")


@pytest.fixture()
def par2():
    return Parameter('size', default="bar", custom=[20])


@pytest.fixture()
def args():
    return ['flow.py', 'run', '--parameter1', 'bar', '--tag', 'ignore']


@pytest.fixture()
def help_args():
    return ['--help', 'flow.py', 'run', '--parameter1', 'bar', '--tag', 'ignore']


@pytest.fixture()
def resume_args():
    return ['flow.py', 'resume', '--tag', 'parameter1:foo', '--tag', 'parameter2:bar']


@pytest.fixture()
def sweep_json():
    return '{"parameter1":[1,2,3],"parameter2":[10,20]}'


@pytest.fixture()
def sweep_json2():
    return '{"parameter1":[-1,-2,-3],"parameter2":[-10,-20]}'


@pytest.fixture()
def sweep_args(sweep_json2):
    return ['flow.py', 'run', '--sweep', sweep_json2, '--tag', 'ignore']


@pytest.fixture()
def par_sweep(sweep_json):
    return Parameter('sweep', default=sweep_json, type=JSONType)


@pytest.fixture()
def cls(par, par_sweep):
    def f(p):
        f.parameter1 = p
        f.sweep = par_sweep
        f._private = 0
        f.not_a_parameter = 0
    f(par)
    return f


@pytest.fixture()
def o(cls):
    out = ParameterizingOut(cls, Parameter, JSONType)
    return out


@pytest.fixture()
def i():
    return In()


@pytest.fixture()
def io():
    return [[In(), Out()]]


@pytest.fixture()
def get_etl(par, par2):
    class Model(ETL):
        etl_parameter = par
        parameter1 = par
        parameter2 = par2
        parameter3 = Parameter('par3', default=42)

        @tim
        def task(self, log_time={}):
            return 42

        def do_sweep(self, i, o):
            ls = list(o.sweep.items())
            sweep1var, sweep1val = ls[o.sweep_index]
            o.sweep_index += 1

            setattr(o, sweep1var, sweep1val[0])

            sweep1 = Sweep(sweep1var, i, o)
            for a in sweep1val:
                i, o = sweep1.add(a)
                o.a = o.parameter1
                o.b = o.etl_parameter
                o.c = o.parameter3
            i, o = sweep1.load()
            sweep1.load_io()

        def load(self, i, o):
            if hasattr(o, 'sweep'):
                self.do_sweep(i, o)
            o.result = self.task()
            o.result = self.task(log_time={})
            super().load(i, o)
    return lambda i, o: Model(i, o)


@pytest.fixture()
def parameterized(get_etl, par):
    class Script(Parameterized):
        factories = {"get_etl": get_etl}
        doc = "A parameterized script"
        _Parameter = Parameter
        _JSONType = JSONType
        parameter1 = par

        def populate(self):
            # get default parameter values from parameterized class
            for var in dir(self):
                if var[0] == '_':
                    continue
                val = getattr(self, var)
                if isinstance(val, self._Parameter):
                    default = val.kwargs['default']
                    setattr(self, var, json.loads(default) if val.type == self._JSONType else default)

    return Script


@pytest.fixture()
def sweeping(parameterized, par_sweep):
    class Sweeping(parameterized):
        sweep = par_sweep

    return Sweeping


def test_can_read_parameter_defaults(o):
    o.read_parameter_defaults()
    assert o.parameter1 == 'foo'
    assert not hasattr(o, 'not_a_parameter')
    assert not hasattr(o, '_private')


def test_can_read_parameter_overrides(o, args):
    o.read_parameter_defaults()
    o.read_parameter_overrides([])
    o.read_parameter_overrides(args)
    assert o.parameter1 == 'bar'
    assert not hasattr(o, 'tag')


def test_can_parse_parameter_sweep(o, sweep_json):
    o.parse_sweep(sweep_json)
    assert o.parameter1 == 1
    assert o.parameter2 == 10
    assert o.parameterized.parameter1 == 'swept'
    assert o.sweep['parameter1'] == [1, 2, 3]
    assert o.sweep['parameter2'] == [10, 20]


def test_can_read_parameter_values(o, args):
    o.read_parameter_values()
    assert o.parameter1 == 1
    assert o.parameter2 == 10
    assert o.parameterized.parameter1 == 'swept'


def test_can_read_sweep_override(o, sweep_args):
    o.read_parameter_values(sweep_args)
    assert o.parameter1 == -1
    assert o.parameter2 == -10
    assert o.parameterized.parameter1 == 'swept'
    assert o.sweep['parameter1'] == [-1, -2, -3]


def test_can_parameterize_help(o, help_args):
    o.read_parameter_defaults()
    o.read_parameter_overrides(help_args)
    assert o.parameter1 == 'bar'
    assert not hasattr(o, 'tag')


def test_can_parameterize_resume(o, resume_args):
    o.read_parameter_defaults()
    o.read_parameter_overrides(resume_args)
    assert o.parameter1 == 'foo'
    assert not hasattr(o, 'tag')


def test_can_inject_parameter_type(par):
    class Foo(Parameter):
        pass
    impl = par.as_injected(Foo)
    assert type(impl) == Foo
    assert impl.name == par.name
    assert impl.default == par.default


def test_can_add_parameter(o, par, par_sweep):
    o.add_parameter('par1', par)
    o.add_parameter('par1', par)
    o.add_parameter('sweep', par_sweep)
    o.add_parameter('sweep', par_sweep)
    assert o.par1 == par.default
    assert o.sweep['parameter1'] == [1, 2, 3]
    assert o.parameterized.par1.default == par.default
    assert o.parameterized.sweep.default == par_sweep.default


def test_can_parameterize(parameterized):
    script = parameterized()
    assert script.parameter1.default == "foo"
    assert script.etl_parameter.default == "foo"


def test_can_parameterize_sweep(sweeping, sweep_json):
    script = sweeping()
    assert script.parameter1 == "swept"
    assert script.sweep.default == sweep_json


def test_can_autotag(parameterized, args):
    script = parameterized()
    script.add_parameter_tags(args=args)
    assert args[-8:] == ['--tag', 'parameter1:bar', '--tag', 'etl_parameter:foo', '--tag', 'parameter2:bar', '--tag', 'parameter3:42']


def test_can_process_input(sweeping, i):
    script = sweeping()
    script.timeit = timeit
    script.populate()
    script.process(i)
    script.etl.et(i, script)
    script.etl.tl(i, script)
    assert script.result == 42


def test_can_process_list(parameterized, io):
    script = parameterized()
    script.process(io)
    assert script.result == 42



import pytest
from paretl import DebugIn, DebugOut


@pytest.fixture()
def i():
    return DebugIn()


@pytest.fixture()
def o():
    return DebugOut()


@pytest.fixture()
def parameter():
    def f(x):
        f.default = x
    f('foo')
    return f


@pytest.fixture()
def logger():
    def f(k, v):
        setattr(f, k, v)
    return f


def test_can_debug_input(i):
    i.foo = 'bar'
    assert i.foo == 'bar'


def test_can_debug_output(o, parameter):
    o.add_parameter('par', parameter)
    # should ignore second call with same parameter name
    o.add_parameter('par', None)
    o.foo = 'bar'
    assert o.foo == 'bar'


def test_can_log_output(logger):
    o = DebugOut(logger=logger)
    o.foo = 'bar'
    assert o.foo == 'bar'
    assert logger.foo == 'bar'


def test_can_hide_output(logger):
    o = DebugOut(logger=logger, hide=['foo'])
    o.foo = 'bar'
    assert o.foo == 'bar'
    assert not hasattr(logger, 'foo')



"""Classes for input and output objects used for passing along state and gathering output in ETLs.

"""


class In:
    """
    Base class for intermediate-state-passing dubbed input, in or i.

    """
    pass


class Out:
    """
    Base class for final-state-passing and output gathering dubbed output, out or o.

    """
    pass


class DebugIn(In):
    """Input class for debugging that does nothing.

    """
    pass


class DebugOut(Out):
    """Output class for debugging that simply stores and logs output.

    """

    def __init__(self, hide=[], logger=lambda k, v: print('o.%s ' % k, '=', v)):
        """

        Arguments:
            hide (list) names of attributes to hide from logging
            logger (lambda k,v) called when an attribute is set
        """
        self.__dict__['data'] = Out()
        self.__dict__['hide'] = hide
        self.__dict__['logger'] = logger
        super().__init__()

    def add_parameter(self, var, val):
        if not hasattr(self, var):
            setattr(self, var, val.default)

    def __setattr__(self, k, v):
        if not k.startswith('_') and k not in self.hide:
            self.logger(k, v)
        self.__dict__[k] = v
        self.data.__dict__[k] = v



"""Classes related to parameter sweeps.

"""


class Result:
    """
    Base class for a result of a parameter sweep.

    """
    def __init__(self, kv={}):
        """

        Arguments:
            kv (dict) initial results
        """
        self.__dict__.update(kv)

    def as_dict(self):
        """
        Method to represent the result as a dict

        Returns:
            the resul as a dict
        """
        return dict(self.__dict__)


class Sweep:
    """
    Base class for a parameter sweep.

    Attributes:
        i (object) the undecorated input
        o (object) the undecorated input
        ires (object) the input result
        ores (object) the output result
    """

    def __init__(self, key, i, o):
        """

        Arguments:
            key (string) name of parameter being swept
            i (object) the input
            o (object) the output
        """
        self.key = key
        self.i = i
        self.o = o
        self.ires = {}
        self.ores = {}

    def add(self, value):
        """
        Method to add a parameter value case to the sweep

        Arguments:
            value (object) the parameter value

        Returns:
            dedicated swept input and output objects to be used in the ETL for this case
        """
        ires = self.ires[value] = Result({self.key: value})
        ores = self.ores[value] = Result({self.key: value})
        i = Swept(self.i, ires)
        o = Swept(self.o, ores)
        return i, o

    def load(self):
        """
        Method to load the swept output result of the sweep into the parent output

        Returns:
            the parent input and output objects
        """
        setattr(self.o, "%s_sweep" % self.key, {k: v.as_dict() for k, v in self.ores.items()})
        return self.i, self.o

    def load_io(self):
        """
        Method to load both the swept input and output result of the sweep into the parent input and output respectively

        Returns:
            the parent input and output objects
        """
        setattr(self.i, "%s_sweep" % self.key, {k: v.as_dict() for k, v in self.ires.items()})
        setattr(self.o, "%s_sweep" % self.key, {k: v.as_dict() for k, v in self.ores.items()})
        return self.i, self.o


class Swept:
    """
    Base class decorating input and output with their swept dittos to be used in ETLs for a case in a parameter sweep.

    Attributes:
        parent (object) the parent input or output
        result (object) the result for gathering results
    """

    def __init__(self, parent, result):
        """

        Arguments:
            parent (object) the parent input or output
            result (Result) the result object for gathering results
        """
        self.__dict__["parent"] = parent
        self.__dict__["result"] = result

    def __setattr__(self, key, value):
        # set attribute both in self and result
        self.__dict__[key] = value
        setattr(self.result, key, value)

    def __getattr__(self, key):
        # get attribute from self, result or parent
        if hasattr(self.result, key):
            return getattr(self.result, key)
        else:
            return getattr(self.parent, key)



"""Base classes for parameterized Extract-Transform-Load in state-passing style.

"""
from .parameter import Parameter


class ETL:
    """
    Base class for parameterized Extract-Transform-Load.
    Input and output objects are used for passing along state and gathering output in ETLs.

    Attributes:
        i (object): The input object supplied at initialization. Intermediate state-passing.
        o (object): The output object supplied at initialization. Final state-passing and output gathering.
    """

    def __init__(self, i, o):
        """

        Args:
            i (object): The input object.
            o (object): The output object.
        """
        self.i = i
        self.o = o

        # add parameters to output if it has that capability
        if hasattr(o, 'add_parameter'):
            self._add_parameters(i, o)

    def _add_parameters(self, i, o):
        """Add own parameters to output.

        Args:
            i (object): The input object.
            o (object): The output object.
        """
        for var, val in self._get_parameters():
            o.add_parameter(var, val)

    def _get_parameters(self):
        """Provide own parameters as a generator of name,value pairs.

        """
        # go through attributes by name
        for name in dir(self):
            # skip private
            if name[0] == '_':
                continue

            # make sure it is accessible and get attribute value
            # try:
            value = getattr(self, name, None)
            # except Exception:
            #     continue

            # if value is a parameter yield name,value
            if isinstance(value, Parameter):
                yield name, value

    def etl(self, i, o):
        """Convenience method for Extract-Transform-Load in one.

        Args:
            i (object): The input object.
            o (object): The output object.
        """
        self.extract(i, o)
        self.transform(i, o)
        self.load(i, o)

    def et(self, i, o):
        """Convenience method for Extract-Transform in one.

        Args:
            i (object): The input object.
            o (object): The output object.
        """
        self.extract(i, o)
        self.transform(i, o)

    def tl(self, i, o):
        """Convenience method for Transform-Load in one.

        Args:
            i (object): The input object.
            o (object): The output object.
        """
        self.transform(i, o)
        self.load(i, o)

    def extract(self, i, o):
        """Placeholder method for Extract that does nothing.

        Args:
            i (object): The input object.
            o (object): The output object.
        """
        pass

    def transform(self, i, o):
        """Placeholder method for Transform that does nothing.

        Args:
            i (object): The input object.
            o (object): The output object.
        """
        pass

    def load(self, i, o):
        """Placeholder method for Load that does nothing.

        Args:
            i (object): The input object.
            o (object): The output object.
        """
        pass
